::xo::library doc {
  SCORM RTE

  This file contains the SCORM Runtime Enviroment implementation:

  * A generic class to implement SCORM data models (like CMI)
  * The CMI Data Model
}

::xo::library require cmi-datatype-procs

namespace eval ::xo {}
namespace eval ::xo::scorm {}
namespace eval ::xo::scorm::data_model {

  namespace import ::xo::scorm::data_model::data_types::*

  Class Element -parameter {
    {readable true}
    {writeable true}
    datatype
    {mandatory true}
    {default_value ""}
  } -ad_doc {
    This class can be used to define SCORM Data Model Elements like
    the "student_id" in "cmi.core.student_id". The implementation of
    these elements may be mandatory or optional for the LMS. The element
    may be read-only, write-only and read/write. Also, the SCORM defines
    allowed datatypes for each element, like CMIIdentifier or CMIString255.
  }

  #
  # String Interface
  #

  Element instproc get_value {} {
    return [[my get_datatype_instance] to_string]
  }

  Element instproc set_value {value} {
    set dt [data_types::[my datatype] new -value $value]
    return [my set_datatype_instance $dt]
  }

  #
  # Typed Interface
  #

  Element instproc set_datatype_instance {datatype_instance} {
    set fp [::xo::scorm::session get_sco_session_data]
    $fp set_property -new true [my formpage_property_name] [$datatype_instance to_string]
    #$fp save
    $fp flush_from_cache_and_refresh
  }

  Element instproc get_datatype_instance {} {
    set fp [::xo::scorm::session get_sco_session_data]
    set prop [$fp get_property -name [my formpage_property_name] -default [my default_value]]
    #my log "Instantiate Datatype [my datatype] with init value $prop (default was [my default_value])"
    return [data_types::[my datatype] new -value $prop]
  }

  Element instproc validate {value} {
    foreach dt [my datatype] {
      if {[$dt is_valid_value $value]} {
        return true
      }
    }
    return false
  }


  #
  # Persistence
  #

  Element instproc formpage_property_name {} {
    set cmi_name [Model get_data_model_element [self]]
    set cmi_form_field_name [regsub -all {\.} $cmi_name _____]
    #my log "$cmi_name ~> $cmi_form_field_name"
    return $cmi_form_field_name
  }



  ###########
  #         #
  # Keyword #
  #         #
  ###########

  Class Keyword -superclass Element -ad_doc {
    Currently, there are only two keywords, namely:
    _children and
    _count
  }

  Keyword ad_instproc set_value {value} {} {
    return 402; #Invalid set value - element is a keyword
  }

  #########
  #       #
  # Group #
  #       #
  #########

  Class Group

  Group instproc init {} {
    my contains {
      ::xo::scorm::data_model::Keyword create _children \
        -datatype CMIString255 \
        -readable true \
        -writeable false \
        -proc get_value {} {
          set group [namespace qualifiers [self]]
          return [join [$group get_elements] ,]
        }
      ::xo::scorm::data_model::Keyword create _count  \
        -datatype CMIInteger \
        -readable true \
        -writeable false \
        -proc get_value {} {
          set group [namespace qualifiers [self]]
          return [llength [$group get_elements]]
        }
    }
  }

  Group ad_instproc -private get_elements {} {
  } {
    set elements [list]
    foreach c [[self] info children] {
      if {![$c istype ::xo::scorm::data_model::Keyword]} {
        lappend elements [namespace tail $c]
      }
    }
    return $elements
  }

  ###########
  #         #
  #  Model  #
  #         #
  ###########

  #maybe ordered composite is enough?
  Class Model -superclass ::xo::tdom::Object -parameter {
      version
  }

  Model instproc init {} {
      my add [::xo::scorm::data_model::Keyword create _version]
  }

  Model proc get_implementing_object {data_model_element} {
      # Convert e.g. cmi.core.student_id to ::xo::scorm::data_model::cmi::core::student_id
      set requested_object "[namespace qualifiers [self]]::[regsub -all {\.} $data_model_element {::}]"
      #my log "$data_model_element"
      return [expr {[my isobject $requested_object] ? $requested_object : ""}]
  }

  Model proc get_data_model_element {implementing_object} {
      # Convert e.g. ::xo::scorm::data_model::cmi::core::student_id to cmi.core.student_id
      set implementing_object [regsub [namespace qualifiers [self]]:: $implementing_object {}]
      return "[regsub -all {::} $implementing_object {.}]"
  }

  Model create cmi -version "1.0" -contains {
      ::xo::scorm::data_model::Group create core -contains {
          ::xo::scorm::data_model::Element create student_id \
            -datatype "CMIIdentifier" \
            -set readable true \
            -set writeable false \
            -proc get_value {} {
              return "[::xo::cc user_id]"
            }
          ::xo::scorm::data_model::Element create student_name \
            -datatype "CMIString255" \
            -set readable true \
            -set writeable false \
            -proc get_value {} {
              acs_user::get -user_id [::xo::cc user_id] -array user
              # This format is required by SCORM
              return "$user(last_name), $user(first_names)"
            }
          ::xo::scorm::data_model::Element create lesson_location \
            -datatype "CMIString255" \
            -writeable true \
            -readable true
          ::xo::scorm::data_model::Element create credit \
            -datatype "Credit" \
            -writeable false \
            -readable true \
            -ad_proc get_value {} {
              Possible values "credit", "no-credit"
            } {
              return "no-credit"
            }
          ::xo::scorm::data_model::Element create lesson_status \
            -datatype "Status" \
            -default_value "not attempted" \
            -readable true \
            -writeable true
          ::xo::scorm::data_model::Element create entry \
              -datatype "Entry" \
              -readable true \
              -writeable false
          ::xo::scorm::data_model::Element create total_time \
              -datatype "CMITimespan" \
              -default_value "0000:00:00.00" \
              -readable true \
              -writeable false \
              -proc get_value {} {
                set tt [my get_datatype_instance]
                set st [session_time get_datatype_instance]
                set temp_tt [$tt add $st]
                return [$temp_tt to_string]
              }
          ::xo::scorm::data_model::Element create exit \
              -datatype "Exit" \
              -readable false \
              -writeable true
          ::xo::scorm::data_model::Element create session_time \
              -datatype "CMITimespan" \
              -default_value "0000:00:00.00" \
              -readable false \
              -writeable true
          ::xo::scorm::data_model::Group create score -contains {
            # TODO: Allowed are CMIDecimal OR CMIBlank
            ::xo::scorm::data_model::Element create raw \
                -datatype CMIDecimal \
                -writeable true
          }
          # Optional Elements
          ::xo::scorm::data_model::Element create comments \
              -mandatory false \
              -readable true \
              -writeable true \
              -datatype "CMIString4096" \
              -proc set_value {value} {
                return "[my get_value] $value"
              }
      }
      ::xo::scorm::data_model::Element create suspend_data \
          -datatype "CMIString4096"
      ::xo::scorm::data_model::Element create launch_data \
          -readable true \
          -writeable false \
          -datatype "CMIString4096" \
          -proc get_value {} {
            #set package_id [::xo::cc package_id]
            #::xo::xo::scorm::Package require $package_id
            #set m [$package_id get_manifest]
            set current_course_id [::xo::scorm::session get_current_course_id]
            set current_sco_id [::xo::scorm::session get_current_sco_id]
            if {$current_course_id ne "" && $current_sco_id ne "" } {
              ::xo::db::CrClass get_instance_from_db -item_id $current_course_id
              set m [$current_course_id get_manifest]
              set i [$m get_element_by_identifier $current_sco_id]
              if {[my isobject ${i}::adlcpdatafromlms::text]} {
                return [${i}::adlcpdatafromlms::text text]
              }
              return ""
            } else {
              ns_log notice "no sco_id ([::xo::cc serialize])"
            }
          }
  }
}

