::xo::library doc {
  IMS CP Procs

  This file implements IMS-CP concepts in the form of classes.
  The IMS-CP concepts (like "item" all have an XML representation).
  Instances of these classes may be constructed by passing a tdom
  xml fragement, or alternatively by setting instance variables via
  parameters.

  As a result, you get aggregate object structures. For example, when
  you build a Manifest object via an existing imsmanifest file, you end
  up with a huge object structure (more or less an object for each XML
  element). Then, you can e.g. render this object structure as XHTML...

  This seemed to be a relatively nice OO approach in the beginning.
  However, I am not sure, whether it is the best idea to do it this way.
  Alternatively, we could create some "Manager" and "Parser" objects, which
  operate on the tdom XML.

}
### TODO - replace traversing with "decoration" process, so that e.g. decoration of the
###        imsfile objects in an composite only happens if a CP is there (and not the manifest alone)
package require tdom
::xo::library require xml-procs
::xo::library require -package xolrn filesystem-procs
namespace eval ::ims {}
namespace eval ::ims::cp {}
namespace eval ::ims::cp::xml {

    proc require_xml {} {
        # Define tdom commmands for easy xml generation

        # FIXME sometimes after server restart these cmds are not available

        dom createNodeCmd cdataNode   cdata
        dom createNodeCmd textNode    t
        dom createNodeCmd commentNode c
        dom createNodeCmd parserNode  x
        dom createNodeCmd piNode      runtime

        dom createNodeCmd elementNode manifest
        dom createNodeCmd elementNode metadata
        dom createNodeCmd elementNode schema
        dom createNodeCmd elementNode schemaversion
        dom createNodeCmd elementNode organizations
        dom createNodeCmd elementNode organization
        dom createNodeCmd elementNode title
        dom createNodeCmd elementNode item
        dom createNodeCmd elementNode resources
        dom createNodeCmd elementNode resource
        dom createNodeCmd elementNode file
        dom createNodeCmd elementNode dependency
    }

}

namespace eval ::ims::cp {

    ######################
    #
    # FILE
    #
    # ::ims::cp::File
    #
    ######################

    # An IMS-File has a href (either relative local - or an url)
    # We distinguis between
    # *) external files, which only have a absolute internet URL and
    # *) internal files, which have an ::xoutil::File associated, which allows for 

    Class File -parameter {
        dom
        href
    }

    File instproc init {} {
        if {[my exists dom]} {
            foreach att [[my set dom] attributes] {
                my set $att [[my set dom] getAttribute $att]
            }
        }
    }

    File instproc asXML {} {
            xml::require_xml
        xml::file -href [my href]
    }

    File instproc get_including_content_package {} {
        set cp ""
        if {[my exists __parent] && [[namespace qualifiers [my set __parent]] istype Resource]} {
            set resource [namespace qualifiers [my set __parent]]
        }
        if {[info exists $resource] && [[namespace qualifiers [namespace qualifiers [namespace qualifiers $resource]]] istype ContentPackage]} {
            set cp [namespace qualifiers [namespace qualifiers [namespace qualifiers $res]]]
        }
        return $cp

    }

    Class InternalFile -superclass File -parameter {
        fileobj
    }

    # TODO replace this with an external manager, that decorates (see above)
    InternalFile instproc attach_fileobj {} {
            # now we MUST have a href, so lets compute the fs-path
            set cp [my get_including_content_package]
            # Its possible, that we just have a manifest, but not a real CP, e.g.
            if {$cp ne ""} {
                set loc [file join [$cp set location] [my set href]]
                my set fileobj [::xoutil::File create [self]::[my autoname file_%04d] -location $loc]
            }
    }



    Class ExternalFile -superclass File

    ExternalFile instproc init {} {
        next
        #TODO we might do url validity check here
    }

    Class PackageInterchangeFile -superclass ::xoutil::ZipFile

    PackageInterchangeFile instproc unpack {} {
        set folder [next]
        $folder class ContentPackage
        return $folder
    }

    ###############################
    #
    #  IMS PACKAGE
    #
    #  ::ims::cp::ContentPackage
    #
    #  We call it "ContentPackage" to avoid confusion with OpenACS/XoWiki Packages
    #
    ###############################


    Class ContentPackage -superclass ::xoutil::Folder -parameter {
        manifest
    }


  #  # There are different ways to contruct a valid ContentPackage Object
  #  # a) "reverse" from an existing pif
  #  #              e.g. when importing
  #  # b) "forward" by providing resources
  #  #              e.g. when exporting
  #  ContentPackage instproc init {} {
  #      next
  #      if {![my exists manifest] && [file exists [file join [my set location] "imsmanifest.xml"]]} {
  #          # IMPORT use case - we parse the already present imsmanifest file
  #          my set manifest [Manifest create [self]::manifest -location [file join [my set location] "imsmanifest.xml"]]
  #      }
  #  }

    ############
    #          #
    # MANIFEST #
    #          #
    ############

    # TODO: Build only one big Composite instead of three (org, meta, res)

    ::xoutil::XMLClass create Manifest -tdom_cmd_name "manifest" -superclass ::xoutil::File -parameter {
        {identifier [namespace tail [self]]}
        {version "IMS CP 1.1.4"}
        metadata
        organizations
        resources
    }


    # There are 3 ways to construct a valid Manifest instance
    # a) by providing a imsmanifest.xml file (e.g. when importing a PIF)
    # b) by providing an in-memory dom document
    # c) by providing resources objects (e.g. when exporting)

    Manifest instproc init {} {

        if {[my exists resources]} {
            Metadata create         [self]::metadata
            Organizations create    [self]::organizations
            Resources create        [self]::resources

            # CREATION use case - we build the manifest ourselves, thus we need at least resources
            foreach r [my set resources] {
                [self]::resources add $r
            }
            foreach o [my set organizations] {
                [self]::organizations add $o
            }
            my set content "[my asXML]"
            next
            my set name "imsmanifest.xml"
        } else {
            # IMPORT use case - we should have a location given
            next
            # Now we should have a valid xml structure as "content" and be able to construct the objects 
            # the other way round
            set doc [dom parse [my set content]]
            set manifest_node [$doc documentElement]

            my set identifier       [$manifest_node getAttribute identifier ""]
            my set version          [$manifest_node getAttribute version ""]
            my set xml:base         [$manifest_node getAttribute xml:base ""]
            my set cwsp:packageType [$manifest_node getAttribute cwsp:packageType ""]

            Metadata create      [self]::metadata       -dom [lindex [$doc getElementsByTagName "metadata"] 0]
            Organizations create [self]::organizations  -dom [lindex [$doc getElementsByTagName "organizations"] 0]
            Resources create     [self]::resources      -dom [lindex [$doc getElementsByTagName "resources"] 0]
            $doc delete
        }

    }


    # This builds an in-memory dom document based on the aggregate objects (resources etc.)
    # and returns it as xml. This is called during construction, and
    # should be called on any later change of the object structure.

    Manifest instproc asXML {} {
        #[self class] require_tdom_cmd
        set document [dom createDocumentNode]
        $document appendFromScript {
        [my tdomcmd] \
            -xmlns "http://www.imsglobal.org/xsd/imscp_v1p1" \
            -xmlns:imsmd "http://www.imsglobal.org/xsd/imsmd_v1p2" \
            -xmlns:xsi "http://www.w3.org/2001/XMLSchema-instance" \
            -xsi:schemaLocation "http://www.imsglobal.org/xsd/imscp_v1p1 http://www.imsglobal.org/xsd/imscp_v1p1.xsd http://www.imsglobal.org/xsd/imsmd_v1p2 http://www.imsglobal.org/xsd/imsmd_v1p2.xsd" \
            -identifier [my identifier]
        }

        [$document documentElement] appendFromScript {
            [self]::metadata asXML
            [self]::organizations asXML
            [self]::resources asXML
        }

        set xml [$document asXML]
        $document delete

        return $xml
    }

    # According to IMSCP, the default orga is the one referenced in the default attribute,
    # or - if this doesnt exist - the first orga.
    # If no orga exists at all, we return the "implicit" organization, derived from the resources

    Manifest instproc get_default_or_implicit_organization {} {
        switch -- [llength [[self]::organizations children]] {
            0 {return [my get_implicit_organization]}
            1 {return [lindex [[self]::organizations children] 0]}
            default {
                if {[[self]::organizations exists "default"]} {
                    # We assume, that the package is consistent, i.e. the id points to an orga
                    return [my get_element_by_identifier "[[self]::organizations default]"]
                } else {
                    return [lindex [[self]::organizations children] 0]
                }
            }
        }
    }

    Manifest instproc get_organization {-identifier} {
        return [my get_element_by_identifier $identifier]
    }

    # returns the object in the object composite that has the given identifier
    Manifest instproc get_element_by_identifier {{-entry_point ""} identifier} {
        if {$entry_point ne ""} {
            if {[$entry_point exists identifier] && [$entry_point identifier] eq $identifier} {
                #my log "[$entry_point set identifier] eq $identifier ..."
                return $entry_point
            }
            foreach c [$entry_point children] {
              #my log "checking $c"
              if {[$c exists identifier] && [$c set identifier] eq $identifier} {
                return $c
              }
              set el [my get_element_by_identifier -entry_point $c $identifier]
              if {$el ne ""} {return $el}
            }
        } else {
          # We start at top, checking orgas, resources,...
          foreach o [[self] info children] {
            set el [my get_element_by_identifier -entry_point $o $identifier]
            if {$el ne ""} {return $el}
          }
        }
    }

    Manifest instproc get_elements_by_identifierref {identifier} {
        # TODO: only resources? is it enough?
        # TODO: go deeper into the children-tree
        set objects [list]
        foreach o [[self] info children] {
            foreach obj [$o children] {
                if {[$obj exists identifierref] && [$obj set identifierref] eq $identifier} {
                    lappend objects $obj
                    return $obj
                }
            }
        }
        return $objects
    }

    ##############
    #            #
    #  METADATA  #
    #            #
    ##############

    ::xoutil::XMLClass Metadata -tdom_cmd_name "metadata" -parameter {
        dom
    }

    #################
    #               #
    # ORGANIZATIONS #
    #               #
    #################

    ::xoutil::XMLClass Organizations -tdom_cmd_name "organizations" -parameter {
        dom
    }

    Organizations instproc init {} {
        if {[my exists dom]} {
            foreach orgNode [[my set dom] childNodes] {
                [self] add [Organization create [self]::[my autoname organization] -dom $orgNode]
            }
            [my set dom] delete
        }
    }

    # TODO Move elsewhere
    Organizations instproc as_widget_options {} {
      set options [list]
      foreach childnode [my children] {
        set option [list [$childnode title] [$childnode set identifier]]
          lappend options $option
      }
      return $options
    }

    Organizations instproc asXHTML {} {
        xml::require_xml
        require_html_procs
        set dropdown [dom createDocumentNode]
        $dropdown appendFromScript {

            html::div -class "ims-cp-organizations" {
                html::h2 {
                    html::t "Available Organizations"
                }
                html::form {
                    html::select -name "organization_id" {
                        foreach childnode [my children] {
                            #TODO: is this dirty?
                            if {[$childnode set identifier] eq [::xo::cc query_parameter organization_id]} {
                                #TODO: For XHTML this selected-notation is correxct, but how to define a valid HTML4 "empty att"
                                html::option -value "[$childnode set identifier]" -selected "selected" {
                                    html::t [$childnode title]
                                }
                            } else {
                                html::option -value "[$childnode set identifier]" {
                                    html::t [$childnode title]
                                }
                            }
                        }
                    }
                    html::button -type "submit" {html::t "Choose"}
                }
            }
        }
        $dropdown asHTML
    }

    ::xoutil::XMLClass Organization -tdom_cmd_name "organization" -parameter {
        dom
        {identifier [namespace tail [self]]}
        {structure "hierarchical"}
    }

    Organization instproc init {} {
        if {[my exists dom]} {
            my instvar dom
            if {[$dom hasAttribute "structure"]} {
                my set structure [$dom getAttribute "structure"]
            }
            if {[$dom hasAttribute "identifier"]} {
                my set identifier [$dom getAttribute "identifier"]
            }
            foreach child [$dom childNodes] {
                #ns_log notice "PROCESSING AN ORG-CHILD"
                #TODO handle metadata
                # we have only either title or item here
                if {[$child nodeName] eq "title"} {
                    #ns_log notice "\t processing title domNode [$child nodeValue]"
                    [self] add [Title create [self]::[my autoname title] -dom $child]
                } elseif {[$child nodeName] eq "item"} {
                    #ns_log notice "\t processing item domNode [$child nodeName]"
                    [self] add [Item create [self]::[my autoname item] -dom $child]
                } else {
                    #my log "Organization child omitted (maybe a comment?): [$child nodeName]"
                }
            }
           #$dom delete
        }
    }

    Organization ad_instproc get_default_entry_item {} {
      @return Returns the entry point item of this organization, i.e. the item to be
      diplayed by default. It only makes sense if this item has an href, thus we go
      down the tree until we find an item that has an href.

      TODO: I *THINK* the standard says that there must be at least one item in an organization
      and that the first item ist the entrypoint. But this has to be verified!!!!
    } {
      foreach item [my all_children_of_type Item] {
        set item_href [$item href]
        if {$item_href ne ""} {
          return $item
        }
      }
    }

    Organization instproc title {} {
        set label ""
        foreach c [my children_of_type Title] {
           append label "[[$c text] text]"
        }
        return [string trimleft $label]
    }

    Organization instproc asXHTML {} {
        require_html_procs
        ds_comment "[::xo::cc serialize]"
        set tree [dom createDocumentNode]
        $tree appendFromScript {
            html::div -id [my identifier] -class "ims-cp-organization" {
                html::h2 {
                    html::t [my title]
                }
                html::ul {
                    foreach childnode [my children_of_type Item] {
                        $childnode asXHTML
                    }
                }
            }
        }
        $tree asXML
    }

    ##################
    #
    # ITEM 
    #
    # An IMS CP Item
    #
    ##################

    # http://www.imsglobal.org/content/packaging/cpv1p1p4/imscp_bindv1p1p4.html#1671565
    # Identifier is required by spec

    # TODO: Item and Organization have duplicate code

    ::xoutil::XMLClass Item -tdom_cmd_name "item" -parameter {
        dom
        {identifier [namespace tail [self]]}
        identifierref
        {isvisible true}
        parameters
    }

    Item instproc init {args} {
      my destroy_on_cleanup
      next
        if {[my exists dom]} {
            my instvar dom
            #ns_log notice "MI: my set identifier [$dom getAttribute identifier]"
            my set identifier [$dom getAttribute "identifier"]
            if {[$dom hasAttribute "identifierref"]} {
                my set identifierref [$dom getAttribute "identifierref"]
            }
            if {[$dom hasAttribute "parameters"]} {
                my set parameters [$dom getAttribute "parameters"]
            }
            foreach childnode [[my set dom] childNodes] {
                #TODO handle metadata
                if {[$childnode nodeName] eq "title"} {
                   # my set title [Title create [self]::title -dom $childnode]
                    [self] add [Title create [self]::[my autoname title] -dom $childnode]
                } elseif {[$childnode nodeName] eq "adlcp:datafromlms"} {
                  #FIXME: SCORM Extenion - refactor out
                  [self] add [AdlcpDataFromLms create [self]::adlcpdatafromlms -dom $childnode]
                } elseif {[$childnode nodeName] eq "item"} {
                    [self] add [Item create [self]::[my autoname item] -dom $childnode]
                } else {
                   # my log "Organization child omitted (maybe a comment?): [$childnode nodeName]"
                }

            }
           #[my set dom] delete
        }
    }

    #TODO: I dont understand the specfication here. why should one item have more titles?
    Item instproc title {} {
        set label ""
        foreach c [my children_of_type Title] {
           append label "[[$c text] text]"
        }
        return [string trimleft $label]
    }

    Item instproc href {} {
      my instvar identifier
      if {![my exists identifierref]} {
        return ""
      }
      set base_href "[::xo::cc url]/"
      set m [my get_including_manifest]
      append base_href "[$m set xml:base]"
      append base_href [[my get_referenced_element] set href]
      append base_href [expr {[my exists parameters] ? [my set parameters] : ""}]
      #set href [export_vars -base $base_href {{xoscorm_sco_id $identifier}}]
      set href $base_href
    }

    Item instproc asXHTML {} {
      my instvar identifier
        require_html_procs
        if {[my isvisible]} {
            html::li -id "ims-cp-id-$identifier" -class "ims-cp-item" {
                if {[my exists identifierref] && [my get_referenced_element] ne "" && [[my get_referenced_element] istype Resource]} {
                    # FIXME - Remove hardcoded frame target
                    html::a -href [my href] -target "xoscorm-player" {
                        html::t [my title]
                    }
                } else {
                    html::t [my title]
                }
                if {[llength [my children_of_type Item]] > 0} {
                    html::ul {
                        foreach child [my children_of_type Item] {
                            $child asXHTML
                        }
                    }
                }
            }
        }
    }

    Class Traverser
    Traverser instproc get_including_manifest {} {
        set parent_obj [self]
        #ns_log notice "SELF is : $parent_obj"
        while {![$parent_obj istype ::ims::cp::Organizations]} {
            #my msg "PARENT is : $parent_obj "
            if {![$parent_obj exists __parent]} break
            set parent_obj [$parent_obj set __parent]
        }
        # now we are at e.g. "mymanifest::organisations
        #my log "Found as Manifest: [namespace qualifiers $parent_obj]"
        set m [namespace qualifiers $parent_obj]
        return $m
    }

    Item instmixin add Traverser
    File instmixin add Traverser

    # http://www.imsglobal.org/content/packaging/cpv1p1p4/imscp_bestv1p1p4.html#1671878
    Item instproc get_referenced_element {} {
        if {[my exists identifierref]} {
            set ref_el [[my get_including_manifest] \
                get_element_by_identifier [my set identifierref]]
            if {$ref_el ne ""} {
              return $ref_el
            } else {
              my msg "Referenced element for [my set identifierref] not found."
            }
        } else {
          error "No identifierref found."
        }
    }


    ####################
    #
    # FIXME: ADLCP:DATAFROMLMS
    #
    #####################

    ::xoutil::XMLClass AdlcpDataFromLms -tdom_cmd_name "adlcp:datafromlms" -tdom_attributes "" -parameter {
        dom
        text
    }

    AdlcpDataFromLms instproc init {args} {
        my set omit_if_no_children true
        if {[my exists dom] && [[my set dom] hasChildNodes]} {
            set data [[[my set dom] firstChild] nodeValue]
            if {$data ne ""} {
                [self] add [::xoutil::TextNode create [self]::text -text $data]
            }
           #[my set dom] delete
        } elseif {[my exists text]} {
           # [self] add [::xoutil::TextNode create [self]::text -text "[my set text]"]
        }
    }

    AdlcpDataFromLms instproc asXHTML {{-wrap ""}} {
        ; # no output
    }



    ####################
    #
    # TITLE
    #
    #####################

    ::xoutil::XMLClass Title -tdom_cmd_name "title" -tdom_attributes "" -parameter {
        dom
        text
    }

    Title instproc init {} {
        my set omit_if_no_children true
        if {[my exists dom] && [[my set dom] hasChildNodes]} {
            set title [[[my set dom] firstChild] nodeValue]
            if {$title ne ""} {
                [self] add [::xoutil::TextNode create [self]::text -text $title]
            }
           #[my set dom] delete
        } elseif {[my exists text]} {
            [self] add [::xoutil::TextNode create [self]::text -text "[my set text]"]
        }
    }
    #TODO title asxml

   # Title instproc asXML {} {
   #     require_html_procs
   #     return [[my tdomcmd] {html::t "[my set value]"}]
   # }

    Title instproc title {} {
        return [[self]::text text]
    }
    Title instproc asXHTML {{-wrap ""}} {
        require_html_procs
        # TODO configurable wrapper
        if {$wrap eq ""} {
               html::t "[[self]::text text]"
        } else {
            html::span -class "ims-cp-title" {
               html::t "[[self]::text text]"
            }
        }
    }
    #############
    #           #
    # RESOURCES #
    #           #
    #############

    ::xoutil::XMLClass Resources -parameter {
        dom
    }

    Resources instproc init {} {
        # We construct from existing DOM
        if {[my exists dom]} {
            foreach c [[my set dom] childNodes] {
                if {[$c nodeName] eq "resource"} {
                    [self] add [Resource create [self]::[my autoname resource] -dom $c]
                } else {
                    #my log "Child of <resources> omitted (maybe a comment?): [$c nodeName]"
                }
            }
            [my set dom] delete
        }
    }


    Resources instproc asXML {} {
        xml::require_xml
        set resources [self]
        xml::resources {
            foreach r [$resources children] {
                 $r asXML
            }
        }
    }

    # RESOURCE

    ::xoutil::XMLClass Resource -tdom_cmd_name "resource" -parameter {
        dom
        {identifier [namespace tail [self]]}
        {type "webcontent"}
        href
        xmlbase
        metadata
        files
        dependencies
    }

    # A simple way to construct a resource is to provide a file.
    # Then, then
    Resource instproc init {} {
        # TODO this is just for my get_attributes - remove this dependency
        my mixin add ::xo::tdom::Object
        ::xo::OrderedComposite create [self]::files
        ::xo::OrderedComposite create [self]::dependencies
        if {[my exists dom]} {
            my instvar dom
            #Import Use Case - We construct from existing DOM
            foreach att [$dom attributes] {
                # distinguish between namespaced and normal
                if {[llength $att] > 1} {
                    my set "[lindex $att 1]:[lindex $att 0]" [$dom getAttributeNS [lindex $att 2] [lindex $att 0]]
                } else {
                    my set "$att" [$dom getAttribute "$att"]
                }
            }
            foreach dep [$dom getElementsByTagName "dependency"] {
                [self]::dependencies add [Dependency create [my autoname dependency_%04d] -dom $dep]
            }
            foreach fil [$dom getElementsByTagName "file"] {
                # TODO Support external ones
                set intfile [InternalFile create [self]::[my autoname imsfile] -dom $fil]
                [self]::files add $intfile
                $intfile attach_fileobj
            }
            [my set dom] delete
        } elseif {![my exists dependencies] && ![my exists files] && ![my exists href] } {
            #TODO throw error - we need at least an href
            
            ns_log error "ERROR in resource creation"
        }

        if {[my exists files]} {
            foreach f [my set files] {
                [self]::files add $f
            }
        }





        # If no files AND no dependencies are given, we create ONE file with the same href
      #  if {[llength [my set files]] eq 0 && [llength [my set dependencies]] eq 0} {
      #      my create_file -href [my href]
      #  }
    }

    Resource instproc destroy {} {
      [self]::dependencies destroy
      [self]::files destroy
      next
    }

    Resource instproc create_file args {
        eval File create [self]::files::[[self] autoname file_%04d] $args
    }


    Resource instproc asXML {} {
        xml::require_xml
        set files_comp [self]::files
        set files_list [$files_comp set __children]
        # If we contain only one single file, then our "entry point" is clear, thus we set
        # the href to the href of this single file.
        if {![my exists href] && [llength $files_list] eq 1 && [[lindex $files_list 0] exists href]} {
            my set href [[lindex $files_list 0] set href]
        }


        [my tdomcmd] [my get_attributes identifier type xmlbase href] {
                foreach f [$files_comp children] {
                    $f asXML
                }
        }
    }

    ::xoutil::XMLClass Dependency -parameter {
        dom
        resource
    }

    Dependency instproc init {} {
        if {[my exists dom]} {
            foreach att [[my set dom] attributes] {
                my set $att [[my set dom] getAttribute $att]
            }
        }
    }

   # Dependency instproc asXML {} {
   #     xml::require_xml
   #     xml::dependency -identifierref "[[my set resource] identifier" {}
   # }


}

::xo::library source_dependent

