::xo::library doc {
  Serverside SCORM API

  This is the script, that represents the serverside SCORM API.
  In our case, the client-side API Adapter forwards all calls to
  the server-side API, which means, that we handle all API Adapter
  functions (like LMSInitialize,...) here.

  This server side API is responsible for producing compliant SCORM
  error codes, etc.

  @see http://adlnet.gov
}

::xo::library require scorm-session-procs
::xo::library require cmi-datamodel-procs

namespace eval ::xo {}
namespace eval ::xo::scorm {

  namespace import ::xo::scorm::data_model::data_types::*

  ###########################
  #                         #
  #   SCORM API Functions   #
  #                         #
  ###########################

  Class ApiFunction -ad_doc {
    For each SCORM API Function an instance of this class exists. These
    objects have the same name as the client side JavaScript functions.
  } -parameter {
    {expect_empty_parameter false}
    {reset_last_error true}
  }

  ApiFunction ad_instproc log_api_call {} {} {
    #ns_log notice "API > [namespace tail [self]] called with parameters: \"[::xo::cc array get queryparm]\""
  }

  ApiFunction ad_instproc response {} {
    Upon each request (coming from the client side API adapter, respectivley the SCO)
    this method is called.
    @author Michael Aram
  } {
    my log_api_call
    my validate_permissions
    my validate_parameter
    my validate_execution_state
    if {![my istype ApiErrorFunction]} {
      # Error Functions do not reset the error code
      ::xo::scorm::session set_property last_error 0
    }
    # If the script wasnt aborted by the validation process,
    # we compute the "normal" answer.
    my compute_answer
    my return_answer
  }

  ApiFunction instproc validate_permissions {} {
    # TODO: Also check permissions (read)
    if {[::xo::cc user_id] <= 0 || [::xo::cc package_id] <= 0} {
      my log "The user is not logged in or we have no package. Cannot initialize the sco."
      my return_error -errornumber 101 ;# 101 = General Exeption
    }
  }

  ApiFunction instproc validate_parameter {} {
    set par [::xo::cc form_parameter parameter]
    if {[my expect_empty_parameter] && $par ne ""} {
      my log "Expected an empty parameter, but got $par"
      my return_error -errornumber 201 ; # 201 = Invalid arg
    }
  }

  ApiFunction ad_instproc validate_execution_state {} {
    Checks the execution state of the users current SCO. It should always be
    initialized, except for calls to LMSInitialize or calls to error_functions.
    Error functions overwrite this behaviour
    @see ::xo::scorm::ApiErrorFunction
  } {
    if {![::xo::scorm::session is_initialized]} {
      # The session must be initialized before any function call may happen
      my log "not initialized"
      my return_error -errornumber 301
    }
  }

  ApiFunction ad_instproc -private return_answer {} {
    Sends the content of the "answer" instance variable to the client.
    @author Michael Aram
  } {
    my log "API < [namespace tail [self]]: '[my set answer]' (Error: '[::xo::scorm::session get_property last_error]')"
    ns_return 200 text/plain [my set answer]
    ad_script_abort
  }

  ApiFunction ad_instproc return_error {
    {-answer false}
    {-errornumber 401}
  } {
    Respond with a SCORM error to the client side API Adapter. The response
    is not an error in terms of HTTP, which means the HTTP status code is "OK".
    It sets SCORMS "Last Error" and returns "false" to the client.
    @errornumber The SCORM 1.2 error code to be returned to the SCO.
    @author Michael Aram
  } {
      #my log "Returning SCORM error: '$errornumber'"
      ::xo::scorm::session set_property last_error $errornumber
      my set answer $answer
      my return_answer
  }


  #######################
  #                     #
  #   Execution State   #
  #                     #
  #######################

  #
  # LMSInitialize
  #
  ApiFunction create LMSInitialize -ad_doc {
    This object represents the implementation of the SCORM 1.2 API function LMSInitialize().
    @author Michael Aram
    @date 2010-02-11
  } -expect_empty_parameter true

  LMSInitialize ad_proc validate_execution_state {} {
    Check, whether we have a valid SCO execution state to initialize the session.
    @date 2010-02-11
  } {
    #my log "Checking status"
    if {[::xo::scorm::session is_initialized]} {
      # Calling LMSInitialize twice is not allowed
      my return_error -errornumber 101; #General Exception
    }
  }

  LMSInitialize ad_proc compute_answer {} {
    In our case, the server-side initialization of a SCORM package means, that we have a
    @date 2010-02-11
  } {
    my set answer true
    set model ::xo::scorm::data_model::cmi
    # Reset session time
    ${model}::core::session_time set_value "0000:00:00.00"
    set lesson_status ${model}::core::lesson_status
    #my log "LMSInitialize: lesson_status: $lesson_status"
    if {[$lesson_status get_value] eq ""} {
      $lesson_status set_value "not attempted"
    }
    switch -- [${model}::core::exit get_value] {
      "" {
        # "exit" has never been set.
        ${model}::core::entry set_value "ab-initio"
      }
      "suspend" {
        ${model}::core::entry set_value "resume"
      }
      default {
        # \"\" (empty) - "logout" - "time-out" and any other value requires:
        my log "'exit' was set to '[${model}::core::exit get_value]', so we reset 'entry'"
        ${model}::core::entry set_value ""
      }
    }
    #ad_set_client_property "scorm_[::xo::cc package_id]" "Execution State" "Initialized"
    ::xo::scorm::session set_property execution_state "initialized"
  }

  #
  # LMSFinish
  #

  ApiFunction create LMSFinish \
      -expect_empty_parameter true

  LMSFinish ad_proc validate_execution_state {} {
    Check, whether we have a valid SCO execution state to finish the session.
    @return boolean
  } {
    if {![::xo::scorm::session is_initialized]} {
      my return_error -errornumber 301; # 301 = "Not initialized"
      #return false
    }
    ::xo::scorm::session set_property execution_state "finished"
  }


  LMSFinish ad_proc compute_answer {} {
    To finish a session, the status has to be initialized
    @author Michael Aram
  } {
    set model ::xo::scorm::data_model::cmi

    # handle time info
    set total_time [${model}::core::total_time get_datatype_instance]
    set session_time [${model}::core::session_time get_datatype_instance]
    $total_time add $session_time
    ${model}::core::total_time set_datatype_instance $total_time

    # set cmi.core.entry
    if {[${model}::core::exit get_value] eq "suspend"} {
      ${model}::core::entry set_value "resume"
    } else {
      ${model}::core::entry set_value ""
    }

    # set cmi.core.lesson_status
    if {[${model}::core::lesson_status get_value] eq "not attempted"} {
      ${model}::core::lesson_status set_value "incomplete"
    }

    ::xo::scorm::session commit
    my set answer true
  }

  #####################
  #                   #
  #   Data Transfer   #
  #                   #
  #####################

  ApiFunction create LMSGetValue

  LMSGetValue ad_proc -private validate_parameter {} {
    @date 2010-02-11
  } {
    set element [::xo::cc form_parameter element]
    #my log "Requested data model element is $element"
  }
  LMSGetValue ad_proc validate_execution_state {} {
    @see ::xo::scorm::ApiErrorFunction
  } {
    if {![::xo::scorm::session is_initialized]} {
      #LMSGetValue must return an empty string here.
      my return_error -errornumber 301 -answer ""
    }
  }

  LMSGetValue ad_proc log_api_call {} {} {
    ns_log notice "API > LMSGetValue(\"[::xo::cc form_parameter element]\") \
        for SCO \"[::xo::cc form_parameter xoscorm_sco_id]\" \
        in Course \"[::xo::cc form_parameter xoscorm_course_id]\""
  }


  LMSGetValue proc compute_answer {} {
    set cmi_element [::xo::cc form_parameter element]
    set element [::xo::scorm::data_model::Model get_implementing_object $cmi_element]
    if {$element ne ""} {
      #my log "Requested Data Model Object \"$element\""
      if {![$element readable]} {
        my return_error -errornumber 404 -answer "" ;# SCORM Error 404 - Element is write only
      }
      my set answer [$element get_value]
      ::xo::scorm::session set_property last_error 0
      #my log "LMSGetValue $element ---> \"[my set answer]\""
    } else {
      # The XOTcl Object Structure couldnt translate the element from
      # the query into an object. This may have several reasons.
      # Respond with "not implemented"

      if {[string match *_children $cmi_element]} {
        # The SCO requested a children list via _children the keyword,
        # but this seems to be an errorneous query or not implemented
        regsub {\._children} $cmi_element "" cmi_base_element
        # Lets check, if we have an implementation of the parent
        set base_element [::xo::scorm::data_model::Model get_implementing_object $cmi_base_element]
        if {$base_element ne ""} {
          my return_error -errornumber 202 -answer "";# 202 = 'Element cannot have children'
        }
        # If the base element doesnt exist, the code below should return 401
      }
      if {[string match *_count $cmi_element]} {
        # The SCO requested a children count list via the _count keyword,
        # but this seems to be an errorneous query.
        regsub {\._count} $cmi_element "" cmi_base_element
        # Lets check, if we have an implementation of the parent
        set base_element [::xo::scorm::data_model::Model get_implementing_object $cmi_base_element]
        if {$base_element ne ""} {
          my return_error -errornumber 203 -answer "";# 203 = 'Element cannot have children'
        }
      }
      #my log "Not implemented: \"$element\”"
      my return_error -errornumber 401 -answer ""
    }
  }

  ApiFunction create LMSSetValue

  LMSSetValue ad_proc log_api_call {} {} {
    ns_log notice "API > LMSSetValue(\"[::xo::cc form_parameter element]\", \
        \"[::xo::cc form_parameter value]\" ) \
        for SCO \"[::xo::cc form_parameter xoscorm_sco_id]\" \
        in Course \"[::xo::cc form_parameter xoscorm_course_id]\""
  }

  LMSSetValue ad_proc -private validate_parameter {} {
    @date 2010-02-11
  } {
    #set element [::xo::cc form_parameter element]
    set element [::xo::scorm::data_model::Model get_implementing_object "[::xo::cc form_parameter element]"]
    set value "[::xo::cc form_parameter value]"
    if {$element ne ""} {
      if {![$element writeable]} {
        my return_error -errornumber 403; # element is readonly
      }
      #my log "testing $element validate $value"
      #my log "== [$element validate $value]"
      if {![$element validate $value]} {
        my return_error -errornumber 405; # Incorrect Datatype
      }
      if {[namespace tail $element] eq "lesson_status"
          && $value eq "not attempted"} {
        my return_error -errornumber 405
      }
    }
    #my log "LMSSetValue '$element' ---> '$value'."
  }

  LMSSetValue proc compute_answer {} {
    set element [::xo::scorm::data_model::Model get_implementing_object "[::xo::cc form_parameter element]"]
    set value "[::xo::cc form_parameter value]"
    if {$element ne ""} {
      my set answer true
      $element set_value $value
      # set code [::xo::scorm::session get_last_error]
      # my log "LMSSetValue \"$element\" ---> \"$value\" (Code: $code)"
      # if {$code eq ""} {
      #   my set answer true
      #   ::xo::scorm::session set_last_error 0
      # } else {
      #   my log "ERROR $element"
      #   my return_error -errornumber $code
      # }
    } else {
      my log "LMSSetValue: Not implemented: \"$element\”"
      my return_error -errornumber 401; # "Not implemented" error
    }
  }

  ApiFunction create LMSCommit          -expect_empty_parameter true

  LMSCommit ad_proc compute_answer {} {
    LMSCommit is currently not necessary in our implementation, as we send each query from the client
    side API adapter directly to the server and handle it. However, in a future - more optimized - version,
    the LMSCommit function could be implemented to gain more performance.
  } {
    ::xo::scorm::session commit
    my set answer true
  }

  ########################
  #                      #
  #   State Management   #
  #                      #
  ########################

  Class ApiErrorFunction -superclass ApiFunction

  ApiErrorFunction ad_instproc -private validate_execution_state {
  } {
    If LMSGetLastError is the first called function, it must return 0.

    @author Michael Aram
    @creation-date 2010-02-10
  } {
    # do nothing;
  }

  ApiErrorFunction ad_instproc validate_parameter {} {
    Validates the parameter given by the SCO. Expects an integer number
    @see ::xo::scorm::ApiFunction
  } {
    set errornumber [::xo::cc form_parameter errornumber]
    #my log "[self] called with errornumber: \"$errornumber\""
    if {![string is integer $errornumber]} {
      my return_error -answer "" -errornumber 201; # invalid arg
    }
  }

  ApiErrorFunction instproc compute_answer {} {
    set errornumber [::xo::cc form_parameter errornumber]
    set errorstring ""
    if {[my exists scorm_errors($errornumber)]} {
      set errorstring [my set scorm_errors($errornumber)]
    }
    my set answer $errorstring
  }

  ApiErrorFunction create LMSGetLastError    -expect_empty_parameter true

  LMSGetLastError proc compute_answer {} {
    if {![::xo::scorm::session is_initialized] &&
         [::xo::scorm::session get_property last_error] eq ""} {
      # We suppose we are the very first called function
      ::xo::scorm::session set_property last_error 0
    }
    my set answer [::xo::scorm::session get_property last_error]
  }


  ApiErrorFunction create LMSGetErrorString

  LMSGetErrorString array set scorm_errors {
    0    "No Error"
    101  "General Exception"
    201  "Invalid argument error"
    202  "Element cannot have children"
    203  "Element not an array - cannot have count"
    301  "Not initialized"
    401  "Not implemented error"
    402  "Invalid set value, element is a keyword"
    403  "Element is read only"
    404  "Element is write only"
    405  "Incorrect Data Type"
  }

  ApiErrorFunction create LMSGetDiagnostic

  LMSGetDiagnostic array set scorm_errors {
    0    "No vendor specific information provided"
    101  "No vendor specific information provided"
    201  "No vendor specific information provided"
    202  "No vendor specific information provided"
    203  "No vendor specific information provided"
    301  "No vendor specific information provided"
    401  "No vendor specific information provided"
    402  "No vendor specific information provided"
    403  "No vendor specific information provided"
    404  "No vendor specific information provided"
    405  "No vendor specific information provided"
  }

}
